"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const core_2 = require("@yarnpkg/core");
const clipanion_1 = require("clipanion");
const get_1 = tslib_1.__importDefault(require("lodash/get"));
const set_1 = tslib_1.__importDefault(require("lodash/set"));
const unset_1 = tslib_1.__importDefault(require("lodash/unset"));
const ModernEngine_1 = require("../ModernEngine");
function formatStackLine(configuration, caller) {
    // TODO: Should this be in formatUtils? Might not be super useful as a core feature...
    const parts = [];
    if (caller.methodName !== null)
        parts.push(core_2.formatUtils.pretty(configuration, caller.methodName, core_2.formatUtils.Type.CODE));
    if (caller.file !== null) {
        const fileParts = [];
        fileParts.push(core_2.formatUtils.pretty(configuration, caller.file, core_2.formatUtils.Type.PATH));
        if (caller.line !== null) {
            fileParts.push(core_2.formatUtils.pretty(configuration, caller.line, core_2.formatUtils.Type.NUMBER));
            if (caller.column !== null) {
                fileParts.push(core_2.formatUtils.pretty(configuration, caller.line, core_2.formatUtils.Type.NUMBER));
            }
        }
        parts.push(`(${fileParts.join(core_2.formatUtils.pretty(configuration, `:`, `grey`))})`);
    }
    return parts.join(` `);
}
// eslint-disable-next-line arca/no-default-export
class ConstraintsCheckCommand extends cli_1.BaseCommand {
    constructor() {
        super(...arguments);
        this.fix = clipanion_1.Option.Boolean(`--fix`, false, {
            description: `Attempt to automatically fix unambiguous issues, following a multi-pass process`,
        });
        this.json = clipanion_1.Option.Boolean(`--json`, false, {
            description: `Format the output as an NDJSON stream`,
        });
    }
    async execute() {
        var _a;
        const configuration = await core_1.Configuration.find(this.context.cwd, this.context.plugins);
        const { project } = await core_1.Project.find(configuration, this.context.cwd);
        const userConfig = await project.loadUserConfig();
        let engine;
        if (userConfig === null || userConfig === void 0 ? void 0 : userConfig.constraints) {
            engine = new ModernEngine_1.ModernEngine(project);
        }
        else {
            const { Constraints } = await Promise.resolve().then(() => tslib_1.__importStar(require(`../Constraints`)));
            engine = await Constraints.find(project);
        }
        let root;
        let hasErrors = false;
        for (let t = 0, T = this.fix ? 10 : 1; t < T; ++t) {
            root = { children: [] };
            const result = await engine.process();
            if (!result)
                break;
            const updates = [];
            for (const [workspaceCwd, workspaceUpdates] of result.manifestUpdates) {
                const workspaceErrors = (_a = result.reportedErrors.get(workspaceCwd)) !== null && _a !== void 0 ? _a : [];
                const workspace = project.getWorkspaceByCwd(workspaceCwd);
                const manifest = workspace.manifest.exportTo({});
                let changedWorkspace = false;
                for (const [fieldPath, newValues] of workspaceUpdates) {
                    if (newValues.size > 1) {
                        const conflictingValuesMessage = [...newValues].map(([value, sources]) => {
                            const prettyValue = core_2.formatUtils.pretty(configuration, value, core_2.formatUtils.Type.INSPECT);
                            const stackLine = sources.size > 0
                                ? formatStackLine(configuration, sources.values().next().value)
                                : null;
                            return stackLine !== null
                                ? `\n${prettyValue} at ${stackLine}`
                                : `\n${prettyValue}`;
                        }).join(``);
                        workspaceErrors.push(`Conflict detected in constraint targeting ${core_2.formatUtils.pretty(configuration, fieldPath, core_2.formatUtils.Type.CODE)}; conflicting values are:${conflictingValuesMessage}`);
                    }
                    else {
                        const [[newValue]] = newValues;
                        const currentValue = (0, get_1.default)(manifest, fieldPath);
                        if (currentValue === newValue)
                            continue;
                        if (!this.fix) {
                            const errorMessage = typeof currentValue === `undefined`
                                ? `Missing field ${core_2.formatUtils.pretty(configuration, fieldPath, core_2.formatUtils.Type.CODE)}; expected ${core_2.formatUtils.pretty(configuration, newValue, core_2.formatUtils.Type.INSPECT)}`
                                : typeof newValue === `undefined`
                                    ? `Extraneous field ${core_2.formatUtils.pretty(configuration, fieldPath, core_2.formatUtils.Type.CODE)} currently set to ${core_2.formatUtils.pretty(configuration, currentValue, core_2.formatUtils.Type.INSPECT)}`
                                    : `Invalid field ${core_2.formatUtils.pretty(configuration, fieldPath, core_2.formatUtils.Type.CODE)}; expected ${core_2.formatUtils.pretty(configuration, newValue, core_2.formatUtils.Type.INSPECT)}, found ${core_2.formatUtils.pretty(configuration, currentValue, core_2.formatUtils.Type.INSPECT)}`;
                            workspaceErrors.push(errorMessage);
                            continue;
                        }
                        if (typeof newValue === `undefined`)
                            (0, unset_1.default)(manifest, fieldPath);
                        else
                            (0, set_1.default)(manifest, fieldPath, newValue);
                        changedWorkspace = true;
                    }
                }
                if (changedWorkspace) {
                    const indent = workspace.manifest.indent;
                    workspace.manifest = new core_1.Manifest();
                    workspace.manifest.indent = indent;
                    workspace.manifest.load(manifest);
                    updates.push(workspace.persistManifest());
                }
                if (workspaceErrors.length > 0) {
                    const errorNodes = [];
                    for (const error of workspaceErrors) {
                        const lines = error.split(/\n/);
                        errorNodes.push({
                            value: core_2.formatUtils.tuple(core_2.formatUtils.Type.NO_HINT, lines[0]),
                            children: lines.slice(1).map(line => ({
                                value: core_2.formatUtils.tuple(core_2.formatUtils.Type.NO_HINT, line),
                            })),
                        });
                    }
                    const workspaceNode = {
                        value: core_2.formatUtils.tuple(core_2.formatUtils.Type.LOCATOR, workspace.anchoredLocator),
                        children: core_1.miscUtils.sortMap(errorNodes, node => node.value[1]),
                    };
                    root.children.push(workspaceNode);
                    hasErrors = true;
                }
            }
            if (updates.length === 0) {
                break;
            }
        }
        root.children = core_1.miscUtils.sortMap(root.children, node => {
            return node.value[1];
        });
        if (hasErrors) {
            core_1.treeUtils.emitTree(root, {
                configuration,
                stdout: this.context.stdout,
                json: this.json,
                separators: 1,
            });
        }
        return hasErrors ? 1 : 0;
    }
}
exports.default = ConstraintsCheckCommand;
ConstraintsCheckCommand.paths = [
    [`constraints`],
];
ConstraintsCheckCommand.usage = clipanion_1.Command.Usage({
    category: `Constraints-related commands`,
    description: `check that the project constraints are met`,
    details: `
      This command will run constraints on your project and emit errors for each one that is found but isn't met. If any error is emitted the process will exit with a non-zero exit code.

      If the \`--fix\` flag is used, Yarn will attempt to automatically fix the issues the best it can, following a multi-pass process (with a maximum of 10 iterations). Some ambiguous patterns cannot be autofixed, in which case you'll have to manually specify the right resolution.

      For more information as to how to write constraints, please consult our dedicated page on our website: https://yarnpkg.com/features/constraints.
    `,
    examples: [[
            `Check that all constraints are satisfied`,
            `yarn constraints`,
        ], [
            `Autofix all unmet constraints`,
            `yarn constraints --fix`,
        ]],
});
